@@ -1,10 +1,10 @@
-/*  Plzip - A parallel compressor compatible with lzip
+/*  Plzip - Parallel compressor compatible with lzip
     Copyright (C) 2009 Laszlo Ersek.
-    Copyright (C) 2009, 2010, 2011, 2012 Antonio Diaz Diaz.
+    Copyright (C) 2009-2014 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -16,7 +16,7 @@
     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 /*
-    Return values: 0 for a normal exit, 1 for environmental problems
+    Exit status: 0 for a normal exit, 1 for environmental problems
     (file not found, invalid flags, I/O errors, etc), 2 to indicate a
     corrupt or invalid input file, 3 for an internal consistency error
     (eg, bug) which caused plzip to panic.
@@ -34,15 +34,34 @@
 #include <string>
 #include <vector>
 #include <fcntl.h>
-#include <inttypes.h>
 #include <pthread.h>
+#include <stdint.h>
 #include <unistd.h>
 #include <utime.h>
 #include <sys/stat.h>
 #include <lzlib.h>
+#if defined(__MSVCRT__)
+#include <io.h>
+#define fchmod(x,y) 0
+#define fchown(x,y,z) 0
+#define strtoull std::strtoul
+#define SIGHUP SIGTERM
+#define S_ISSOCK(x) 0
+#define S_IRGRP 0
+#define S_IWGRP 0
+#define S_IROTH 0
+#define S_IWOTH 0
+#endif
+#if defined(__OS2__)
+#include <io.h>
+#endif
 
 #include "arg_parser.h"
-#include "plzip.h"
+#include "lzip.h"
+
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
 
 #if CHAR_BIT != 8
 #error "Environments where CHAR_BIT != 8 are not supported."
@@ -53,15 +72,9 @@ namespace {
 
 const char * const Program_name = "Plzip";
 const char * const program_name = "plzip";
-const char * const program_year = "2012";
+const char * const program_year = "2014";
 const char * invocation_name = 0;
 
-#ifdef O_BINARY
-const int o_binary = O_BINARY;
-#else
-const int o_binary = 0;
-#endif
-
 struct { const char * from; const char * to; } const known_extensions[] = {
   { ".lz",  ""     },
   { ".tlz", ".tar" },
@@ -69,8 +82,8 @@ struct { const char * from; const char * to; } const known_extensions[] = {
 
 struct Lzma_options
   {
-  int dictionary_size;		// 4KiB..512MiB
-  int match_len_limit;		// 5..273
+  int dictionary_size;		// 4 KiB .. 512 MiB
+  int match_len_limit;		// 5 .. 273
   };
 
 enum Mode { m_compress, m_decompress, m_test };
@@ -81,33 +94,31 @@ const mode_t usr_rw = S_IRUSR | S_IWUSR;
 const mode_t all_rw = usr_rw | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
 mode_t outfd_mode = usr_rw;
 bool delete_output_on_interrupt = false;
-pthread_t main_thread;
-pid_t main_thread_pid;
 
 
-void show_help()
+void show_help( const long num_online )
   {
-  std::printf( "%s - A parallel compressor compatible with lzip.\n", Program_name );
+  std::printf( "%s - Parallel compressor compatible with lzip.\n", Program_name );
   std::printf( "\nUsage: %s [options] [files]\n", invocation_name );
   std::printf( "\nOptions:\n"
                "  -h, --help                     display this help and exit\n"
                "  -V, --version                  output version information and exit\n"
-               "  -B, --data-size=<bytes>        set input data block size in bytes\n"
+               "  -B, --data-size=<bytes>        set size of input data blocks, in bytes\n"
                "  -c, --stdout                   send output to standard output\n"
                "  -d, --decompress               decompress\n"
                "  -f, --force                    overwrite existing output files\n"
                "  -F, --recompress               force recompression of compressed files\n"
                "  -k, --keep                     keep (don't delete) input files\n"
                "  -m, --match-length=<bytes>     set match length limit in bytes [36]\n"
-               "  -n, --threads=<n>              set the number of (de)compression threads\n"
+               "  -n, --threads=<n>              set number of (de)compression threads [%ld]\n"
                "  -o, --output=<file>            if reading stdin, place the output into <file>\n"
                "  -q, --quiet                    suppress all messages\n"
-               "  -s, --dictionary-size=<bytes>  set dictionary size limit in bytes [8MiB]\n"
+               "  -s, --dictionary-size=<bytes>  set dictionary size limit in bytes [8 MiB]\n"
                "  -t, --test                     test compressed file integrity\n"
                "  -v, --verbose                  be verbose (a 2nd -v gives more)\n"
                "  -1 .. -9                       set compression level [default 6]\n"
                "      --fast                     alias for -1\n"
-               "      --best                     alias for -9\n" );
+               "      --best                     alias for -9\n", num_online );
   if( verbosity > 0 )
     {
     std::printf( "  -D, --debug=<level>        (0-1) print debug statistics to stderr\n" );
@@ -119,7 +130,12 @@ void show_help()
                "The bidimensional parameter space of LZMA can't be mapped to a linear\n"
                "scale optimal for all files. If your files are large, very repetitive,\n"
                "etc, you may need to use the --match-length and --dictionary-size\n"
-               "options directly to achieve optimal performance.\n"
+               "options directly to achieve optimal performance. For example, -9m64\n"
+               "usually compresses executables more (and faster) than -9.\n"
+               "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
+               "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
+               "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
+               "caused plzip to panic.\n"
                "\nReport bugs to lzip-bug@nongnu.org\n"
                "Plzip home page: http://www.nongnu.org/lzip/plzip.html\n" );
   }
@@ -127,23 +143,23 @@ void show_help()
 
 void show_version()
   {
-  std::printf( "%s %s\n", Program_name, PROGVERSION );
+  std::printf( "%s %s\n", program_name, PROGVERSION );
   std::printf( "Copyright (C) 2009 Laszlo Ersek.\n"
                "Copyright (C) %s Antonio Diaz Diaz.\n", program_year );
   std::printf( "Using Lzlib %s\n", LZ_version() );
-  std::printf( "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n"
+  std::printf( "License GPLv2+: GNU GPL version 2 or later <http://gnu.org/licenses/gpl.html>\n"
                "This is free software: you are free to change and redistribute it.\n"
                "There is NO WARRANTY, to the extent permitted by law.\n" );
   }
 
 
-long long getnum( const char * const ptr,
-                  const long long llimit = LLONG_MIN + 1,
-                  const long long ulimit = LLONG_MAX )
+unsigned long long getnum( const char * const ptr,
+                           const unsigned long long llimit,
+                           const unsigned long long ulimit )
   {
+  char * tail;
   errno = 0;
-  char *tail;
-  long long result = strtoll( ptr, &tail, 0 );
+  unsigned long long result = strtoull( ptr, &tail, 0 );
   if( tail == ptr )
     {
     show_error( "Bad or missing numerical argument.", 0, true );
@@ -178,7 +194,7 @@ long long getnum( const char * const ptr,
       }
     for( int i = 0; i < exponent; ++i )
       {
-      if( LLONG_MAX / factor >= llabs( result ) ) result *= factor;
+      if( ulimit / factor >= result ) result *= factor;
       else { errno = ERANGE; break; }
       }
     }
@@ -194,7 +210,7 @@ long long getnum( const char * const ptr,
 
 int get_dict_size( const char * const arg )
   {
-  char *tail;
+  char * tail;
   int bits = std::strtol( arg, &tail, 0 );
   if( bits >= LZ_min_dictionary_bits() &&
       bits <= LZ_max_dictionary_bits() && *tail == 0 )
@@ -216,7 +232,7 @@ int extension_index( const std::string & name )
   }
 
 
-int open_instream( const std::string & name, struct stat * const in_statsp,
+int open_instream( const char * const name, struct stat * const in_statsp,
                    const Mode program_mode, const int eindex,
                    const bool recompress, const bool to_stdout )
   {
@@ -225,31 +241,31 @@ int open_instream( const std::string & name, struct stat * const in_statsp,
     {
     if( verbosity >= 0 )
       std::fprintf( stderr, "%s: Input file '%s' already has '%s' suffix.\n",
-                    program_name, name.c_str(),
-                    known_extensions[eindex].from );
+                    program_name, name, known_extensions[eindex].from );
     }
   else
     {
-    infd = open( name.c_str(), O_RDONLY | o_binary );
+    infd = open( name, O_RDONLY | O_BINARY );
     if( infd < 0 )
       {
       if( verbosity >= 0 )
         std::fprintf( stderr, "%s: Can't open input file '%s': %s.\n",
-                      program_name, name.c_str(), std::strerror( errno ) );
+                      program_name, name, std::strerror( errno ) );
       }
     else
       {
       const int i = fstat( infd, in_statsp );
-      const mode_t & mode = in_statsp->st_mode;
+      const mode_t mode = in_statsp->st_mode;
       const bool can_read = ( i == 0 &&
                               ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
                                 S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
-      if( i != 0 || ( !S_ISREG( mode ) && ( !to_stdout || !can_read ) ) )
+      const bool no_ofile = to_stdout || program_mode == m_test;
+      if( i != 0 || ( !S_ISREG( mode ) && ( !can_read || !no_ofile ) ) )
         {
         if( verbosity >= 0 )
           std::fprintf( stderr, "%s: Input file '%s' is not a regular file%s.\n",
-                        program_name, name.c_str(),
-                        ( can_read && !to_stdout ) ?
+                        program_name, name,
+                        ( can_read && !no_ofile ) ?
                         " and '--stdout' was not specified" : "" );
         close( infd );
         infd = -1;
@@ -288,7 +304,7 @@ void set_d_outname( const std::string & name, const int i )
 
 bool open_outstream( const bool force )
   {
-  int flags = O_CREAT | O_WRONLY | o_binary;
+  int flags = O_CREAT | O_WRONLY | O_BINARY;
   if( force ) flags |= O_TRUNC; else flags |= O_EXCL;
 
   outfd = open( output_filename.c_str(), flags, outfd_mode );
@@ -322,32 +338,20 @@ bool check_tty( const int infd, const Mode program_mode )
   }
 
 
-void cleanup_and_fail( const int retval )
-  {
-  if( delete_output_on_interrupt )
-    {
-    delete_output_on_interrupt = false;
-    if( verbosity >= 0 )
-      std::fprintf( stderr, "%s: Deleting output file '%s', if it exists.\n",
-                    program_name, output_filename.c_str() );
-    if( outfd >= 0 ) { close( outfd ); outfd = -1; }
-    if( std::remove( output_filename.c_str() ) != 0 && errno != ENOENT )
-      show_error( "WARNING: deletion of output file (apparently) failed." );
-    }
-  std::exit( retval );
-  }
-
-
      // Set permissions, owner and times.
 void close_and_set_permissions( const struct stat * const in_statsp )
   {
   bool warning = false;
   if( in_statsp )
     {
+    const mode_t mode = in_statsp->st_mode;
     // fchown will in many cases return with EPERM, which can be safely ignored.
-    if( ( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) != 0 &&
-          errno != EPERM ) ||
-        fchmod( outfd, in_statsp->st_mode ) != 0 ) warning = true;
+    if( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) == 0 )
+      { if( fchmod( outfd, mode ) != 0 ) warning = true; }
+    else
+      if( errno != EPERM ||
+          fchmod( outfd, mode & ~( S_ISUID | S_ISGID | S_ISVTX ) ) != 0 )
+        warning = true;
     }
   if( close( outfd ) != 0 ) cleanup_and_fail( 1 );
   outfd = -1;
@@ -364,12 +368,9 @@ void close_and_set_permissions( const struct stat * const in_statsp )
   }
 
 
-extern "C" void signal_handler( int sig )
+extern "C" void signal_handler( int )
   {
-  if( !pthread_equal( pthread_self(), main_thread ) )
-    kill( main_thread_pid, sig );
-  if( sig != SIGUSR1 )
-    show_error( "Control-C or similar caught, quitting." );
+  show_error( "Control-C or similar caught, quitting." );
   cleanup_and_fail( 1 );
   }
 
@@ -387,13 +388,6 @@ void set_signals()
 int verbosity = 0;
 
 
-// This can be called from any thread, main thread or sub-threads alike,
-// since they all call common helper functions that call fatal() in case
-// of an error.
-//
-void fatal() { signal_handler( SIGUSR1 ); }
-
-
 void Pretty_print::operator()( const char * const msg ) const
   {
   if( verbosity >= 0 )
@@ -402,11 +396,11 @@ void Pretty_print::operator()( const char * const msg ) const
       {
       first_post = false;
       std::fprintf( stderr, "  %s: ", name_.c_str() );
-      for( unsigned int i = 0; i < longest_name - name_.size(); ++i )
+      for( unsigned i = 0; i < longest_name - name_.size(); ++i )
         std::fprintf( stderr, " " );
       if( !msg ) std::fflush( stderr );
       }
-    if( msg ) std::fprintf( stderr, "%s.\n", msg );
+    if( msg ) std::fprintf( stderr, "%s\n", msg );
     }
   }
 
@@ -419,10 +413,10 @@ void show_error( const char * const msg, const int errcode, const bool help )
       {
       std::fprintf( stderr, "%s: %s", program_name, msg );
       if( errcode > 0 )
-        std::fprintf( stderr, ": %s", std::strerror( errcode ) );
+        std::fprintf( stderr, ": %s.", std::strerror( errcode ) );
       std::fprintf( stderr, "\n" );
       }
-    if( help && invocation_name && invocation_name[0] )
+    if( help )
       std::fprintf( stderr, "Try '%s --help' for more information.\n",
                     invocation_name );
     }
@@ -432,15 +426,63 @@ void show_error( const char * const msg, const int errcode, const bool help )
 void internal_error( const char * const msg )
   {
   if( verbosity >= 0 )
-    std::fprintf( stderr, "%s: internal error: %s.\n", program_name, msg );
+    std::fprintf( stderr, "%s: internal error: %s\n", program_name, msg );
   std::exit( 3 );
   }
 
 
+// This can be called from any thread, main thread or sub-threads alike,
+// since they all call common helper functions that call cleanup_and_fail()
+// in case of an error.
+//
+void cleanup_and_fail( const int retval )
+  {
+  // only one thread can delete and exit
+  static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
+
+  pthread_mutex_lock( &mutex );		// ignore errors to avoid loop
+  if( delete_output_on_interrupt )
+    {
+    delete_output_on_interrupt = false;
+    if( verbosity >= 0 )
+      std::fprintf( stderr, "%s: Deleting output file '%s', if it exists.\n",
+                    program_name, output_filename.c_str() );
+    if( outfd >= 0 ) { close( outfd ); outfd = -1; }
+    if( std::remove( output_filename.c_str() ) != 0 && errno != ENOENT )
+      show_error( "WARNING: deletion of output file (apparently) failed." );
+    }
+  std::exit( retval );
+  }
+
+
+void show_progress( const int packet_size,
+                    const Pretty_print * const p,
+                    const unsigned long long cfile_size )
+  {
+  static unsigned long long csize = 0;		// file_size / 100
+  static unsigned long long pos = 0;
+  static const Pretty_print * pp = 0;
+  static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
+
+  if( p )					// initialize static vars
+    { csize = cfile_size; pos = 0; pp = p; }
+  if( pp )
+    {
+    xlock( &mutex );
+    pos += packet_size;
+    if( csize > 0 )
+      std::fprintf( stderr, "%4llu%%", pos / csize );
+    std::fprintf( stderr, "  %.1f MB\r", pos / 1000000.0 );
+    pp->reset(); (*pp)();			// restore cursor position
+    xunlock( &mutex );
+    }
+  }
+
+
 int main( const int argc, const char * const argv[] )
   {
-  // Mapping from gzip/bzip2 style 1..9 compression modes
-  // to the corresponding LZMA compression modes.
+  /* Mapping from gzip/bzip2 style 1..9 compression modes
+     to the corresponding LZMA compression modes. */
   const Lzma_options option_mapping[] =
     {
     { 1 << 20,   5 },		// -0
@@ -454,6 +496,9 @@ int main( const int argc, const char * const argv[] )
     { 3 << 23, 132 },		// -8
     { 1 << 25, 273 } };		// -9
   Lzma_options encoder_options = option_mapping[6];	// default = "-6"
+  std::string input_filename;
+  std::string default_output_filename;
+  std::vector< std::string > filenames;
   int data_size = 0;
   int debug_level = 0;
   int infd = -1;
@@ -463,16 +508,12 @@ int main( const int argc, const char * const argv[] )
   bool keep_input_files = false;
   bool recompress = false;
   bool to_stdout = false;
-  std::string input_filename;
-  std::string default_output_filename;
-  std::vector< std::string > filenames;
   invocation_name = argv[0];
-  main_thread = pthread_self();
-  main_thread_pid = getpid();
 
   if( LZ_version()[0] != LZ_version_string[0] )
-    internal_error( "bad library version" );
+    internal_error( "bad library version." );
 
+  const long num_online = std::max( 1L, sysconf( _SC_NPROCESSORS_ONLN ) );
   long max_workers = sysconf( _SC_THREAD_THREADS_MAX );
   if( max_workers < 1 || max_workers > INT_MAX / (int)sizeof (pthread_t) )
     max_workers = INT_MAX / sizeof (pthread_t);
@@ -532,13 +573,13 @@ int main( const int argc, const char * const argv[] )
       case 'D': debug_level = getnum( arg, 0, 3 ); break;
       case 'f': force = true; break;
       case 'F': recompress = true; break;
-      case 'h': show_help(); return 0;
+      case 'h': show_help( num_online ); return 0;
       case 'k': keep_input_files = true; break;
       case 'm': encoder_options.match_len_limit =
                   getnum( arg, LZ_min_match_len_limit(),
                                LZ_max_match_len_limit() ); break;
-      case 'o': default_output_filename = arg; break;
       case 'n': num_workers = getnum( arg, 1, max_workers ); break;
+      case 'o': default_output_filename = arg; break;
       case 'q': verbosity = -1; break;
       case 's': encoder_options.dictionary_size = get_dict_size( arg );
                 break;
@@ -546,13 +587,13 @@ int main( const int argc, const char * const argv[] )
       case 't': program_mode = m_test; break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
-      default : internal_error( "uncaught option" );
+      default : internal_error( "uncaught option." );
       }
     } // end process options
 
-#if defined(__OS2__)
-  _fsetmode( stdin, "b" );
-  _fsetmode( stdout, "b" );
+#if defined(__MSVCRT__) || defined(__OS2__)
+  setmode( STDIN_FILENO, O_BINARY );
+  setmode( STDOUT_FILENO, O_BINARY );
 #endif
 
   if( program_mode == m_test )
@@ -561,43 +602,39 @@ int main( const int argc, const char * const argv[] )
   if( data_size <= 0 )
     data_size = 2 * std::max( 65536, encoder_options.dictionary_size );
   else if( data_size < encoder_options.dictionary_size )
-    encoder_options.dictionary_size = std::max( data_size, LZ_min_dictionary_size() );
+    encoder_options.dictionary_size =
+      std::max( data_size, LZ_min_dictionary_size() );
 
   if( num_workers <= 0 )
-    {
-    long num_online = sysconf( _SC_NPROCESSORS_ONLN );
-    if( num_online <= 0 ) num_online = 1;
     num_workers = std::min( num_online, max_workers );
-    }
 
   bool filenames_given = false;
   for( ; argind < parser.arguments(); ++argind )
     {
-    if( parser.argument( argind ) != "-" ) filenames_given = true;
     filenames.push_back( parser.argument( argind ) );
+    if( filenames.back() != "-" ) filenames_given = true;
     }
 
   if( filenames.empty() ) filenames.push_back("-");
   if( !to_stdout && program_mode != m_test &&
       ( filenames_given || default_output_filename.size() ) )
     set_signals();
-  std::signal( SIGUSR1, signal_handler );
 
   Pretty_print pp( filenames );
 
   int retval = 0;
-  for( unsigned int i = 0; i < filenames.size(); ++i )
+  for( unsigned i = 0; i < filenames.size(); ++i )
     {
     struct stat in_stats;
     output_filename.clear();
 
-    if( !filenames[i].size() || filenames[i] == "-" )
+    if( filenames[i].empty() || filenames[i] == "-" )
       {
       input_filename.clear();
       infd = STDIN_FILENO;
       if( program_mode != m_test )
         {
-        if( to_stdout || !default_output_filename.size() )
+        if( to_stdout || default_output_filename.empty() )
           outfd = STDOUT_FILENO;
         else
           {
@@ -607,7 +644,7 @@ int main( const int argc, const char * const argv[] )
           outfd_mode = all_rw;
           if( !open_outstream( force ) )
             {
-            if( outfd == -1 && retval < 1 ) retval = 1;
+            if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
             continue;
             }
@@ -618,7 +655,7 @@ int main( const int argc, const char * const argv[] )
       {
       input_filename = filenames[i];
       const int eindex = extension_index( input_filename );
-      infd = open_instream( input_filename, &in_stats, program_mode,
+      infd = open_instream( input_filename.c_str(), &in_stats, program_mode,
                             eindex, recompress, to_stdout );
       if( infd < 0 ) { if( retval < 1 ) retval = 1; continue; }
       if( program_mode != m_test )
@@ -632,7 +669,7 @@ int main( const int argc, const char * const argv[] )
           outfd_mode = usr_rw;
           if( !open_outstream( force ) )
             {
-            if( outfd == -1 && retval < 1 ) retval = 1;
+            if( retval < 1 ) retval = 1;
             close( infd ); infd = -1;
             continue;
             }
@@ -645,16 +682,21 @@ int main( const int argc, const char * const argv[] )
     if( output_filename.size() && !to_stdout && program_mode != m_test )
       delete_output_on_interrupt = true;
     const struct stat * const in_statsp = input_filename.size() ? &in_stats : 0;
+    const bool infd_isreg = in_statsp && S_ISREG( in_statsp->st_mode );
     pp.set_name( input_filename );
     if( verbosity >= 1 ) pp();
-    int tmp = 0;
+    int tmp;
     if( program_mode == m_compress )
+      {
+      if( verbosity >= 2 )	// init
+        show_progress( 0, &pp, infd_isreg ? in_statsp->st_size / 100 : 0 );
       tmp = compress( data_size, encoder_options.dictionary_size,
                       encoder_options.match_len_limit,
                       num_workers, infd, outfd, pp, debug_level );
+      }
     else
       tmp = decompress( num_workers, infd, outfd, pp, debug_level,
-                        program_mode == m_test );
+                        program_mode == m_test, infd_isreg );
     if( tmp > retval ) retval = tmp;
     if( tmp && program_mode != m_test ) cleanup_and_fail( retval );
 
